import html.parser
import re
from typing import Optional, Union

from easy_scraper import entity


class HTMLParser(html.parser.HTMLParser):
    def text(self, data: Optional[str]) -> entity.Html:
        if data is None:
            return entity.Empty()
        data = data.strip()
        if self.is_pattern:
            match = re.match(r"{{\s*(\w*)\s*}}", data)
            if match:
                name = match[1]
                return entity.Pattern(name)
        return entity.PlainText(data)

    def __call__(self, data: str, is_pattern: bool = False) -> entity.Elem:
        root = entity.Elem("Root", {}, [], None)
        self.cur = root
        self.is_pattern = is_pattern
        self.feed(data)
        return root

    def handle_starttag(self, tag, attrs):
        attrs = {name: self.text(data) for name, data in attrs}
        child = entity.Elem(tag, attrs, [], self.cur)
        self.cur.children.append(child)
        self.cur = child

    def handle_endtag(self, tag):
        self.cur = self.cur.parent

    def handle_data(self, data: str):
        data = data.strip()
        if not data:
            return
        child = self.text(data)
        self.cur.children.append(child)



from copy import deepcopy
from typing import Optional

from easy_scraper import entity


def match_attr(x: dict, y: dict) -> Optional[dict]:
    """Matching dicts

    Returns
    -------
    Pattern matched result.
    None if failed.
    """
    ret = {}
    for k in y.keys():
        if k not in x:
            return None
        a = x[k]
        b = y[k]
        if isinstance(a, entity.Empty) and not isinstance(b, entity.Empty):
            return None
        if isinstance(a, entity.PlainText) and isinstance(b, entity.PlainText):
            if a.data != b.data:
                return None
        if isinstance(a, entity.PlainText) and isinstance(b, entity.Pattern):
            ret[b.name] = a.data
    return ret


def match_siblings(ref_children: list[entity.Html], pat_children: list[entity.Html]) -> list[dict]:
    if len(ref_children) < len(pat_children):
        return []
    if len(pat_children) == 0:
        return [{}]
    ret = []
    # matching head
    head = match_html(ref_children[0], pat_children[0])
    tail = match_siblings(ref_children[1:], pat_children[1:])
    for r1 in head:
        for r2 in tail:
            ret.append(r1 | r2)
    # skip head
    ret.extend(match_siblings(ref_children[1:], pat_children))
    return ret


def match_html(ref: entity.Html, pattern: entity.Html) -> list[dict[str, str]]:
    """Matching Htmls

    Returns
    -------
    All results.
    Empty if nothing is matched.
    """
    if isinstance(ref, entity.PlainText) and isinstance(pattern, entity.PlainText):
        if ref.data == pattern.data:
            return [{}]
        else:
            return []

    elif isinstance(ref, entity.PlainText) and isinstance(pattern, entity.Pattern):
        return [{pattern.name: ref.data}]

    elif isinstance(ref, entity.Elem) and isinstance(pattern, entity.Elem):
        res = []
        # skip root of ref
        for child in ref.children:
            res.extend(match_html(child, pattern))
        # matching root
        if ref.tag == pattern.tag:
            res_attrs = match_attr(ref.attrs, pattern.attrs)
            if res_attrs is not None:
                for res_siblings in match_siblings(ref.children, pattern.children):
                    res.append(res_attrs | res_siblings)
        return res

    elif isinstance(ref, entity.Elem):
        res = []
        for child in ref.children:
            res.extend(match_html(child, pattern))
        return res

    else:
        return []



__all__ = ["match"]

from easy_scraper.htmlparser import HTMLParser
from easy_scraper.matching import match_html


def match(
    target: str,
    pattern: str,
) -> list[dict[str, str]]:
    """HTML Pattern Matching"""
    parser = HTMLParser()
    ref = parser(target)
    pat = parser(pattern, is_pattern=True)
    return match_html(ref, pat)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['easy_scraper', 'easy_scraper.entity']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'easy-scraper-py',
    'version': '0.1.5',
    'description': 'An Easy Scraper for HTML',
    'long_description': '# easy-scraper-py\n\n![](https://img.shields.io/static/v1?label=+&message=Python%203.9%2B&color=lightblue&logo=Python)\n![](https://img.shields.io/static/v1?label=status&message=Work%20In%20Progress&color=red)\n[![PyPI](https://img.shields.io/pypi/v/easy-scraper-py.svg)](https://pypi.python.org/pypi/easy-scraper-py)\n\nAn easy scraping tool for HTML\n\n## Goal\n\nRe-implementation of [tanakh/easy-scraper](https://github.com/tanakh/easy-scraper) in Python.\n\n## Install from PyPI\n\n```bash\n   pip install easy-scraper-py\n```\n\n## Usage Example\n\n```html\n<!-- Target -->\n<body>\n    <b>NotMe</b>\n    <a class=here>Here</a>\n    <a class=nothere>NotHere</a>\n</body>\n\n<!-- Pattern -->\n<a class=here>{{ text }}</a>\n```\n\n```python\nimport easy_scraper\n\ntarget = r"""<body>\n    <b>NotMe</b>\n    <a class=here>Here</a>\n    <a class=nothere>NotHere</a>\n</body>\n"""  # newlines and spaces are all ignored.\n\npattern = "<a class=here>{{ text }}</a>"\n\neasy_scraper.match(target, pattern)  # [{\'text\': \'Here\'}]\n```\n\n```python\n# XML (RSS) scraping\nimport easy_scraper\nimport urllib.request\n\nbody = urllib.request.urlopen("https://kuragebunch.com/rss/series/10834108156628842505").read().decode()\nres = easy_scraper.match(body, "<item><title>{{ title }}</title><link>{{ link }}</link></item>")\nfor item in res[:5]:\n    print(item)\n```\n',
    'author': 'cympfh',
    'author_email': 'cympfh@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cympfh/easy-scraper-py/',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Union


@dataclass
class Elem:
    tag: str
    attrs: dict[str, str]
    children: list[Html]
    parent: Optional[Html]

    def __repr__(self) -> str:
        return f"<{self.tag} {self.attrs}>{ ' '.join(repr(c) for c in self.children) }</>"


@dataclass
class PlainText:
    data: str


@dataclass
class Pattern:
    name: str


class Empty:
    pass


Html = Union[Elem, PlainText, Pattern, Empty]
